/* ownclass.c (c) 1990 by Zortech - used by permission
   
   You may distribute or use this file without restriction except that
   the above copyright notice should remain in the source file if it is
   electronically or otherwise distributed.
   
   Zortech ltd and Zortech Inc do not warranty that this software will 
   perform according to the descriptions, and will not be responsible for
   any loss or damage consequential or otherwise, which may occur as a
   result of using these functions.
   
   This file contains a number of functions which allow the un-mangling
   of C++ mangled names as used in Zortech C++ v2. It is provided as an
   alternative to unmangle.c since it ignores function parameter information.
   This is useful for debuggers which will normally obtain this information
   from the debug info in the executable file.
   
   Be warned these routines have not been thoroughly tested and may contain
   obscure bugs.
   
   written by Phil Murray and Dave Mansell of Zortech - March 1990
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>


typedef struct unmangle_s {
	char	*name;
	char	*oper;
} UNMANGLE;

static UNMANGLE	um_table[] = {
	{"ct","1"},
	{"dt","2"},
	{"nw"," new"},
	{"dl"," delete"},
	{"pl","+"},
	{"mi","-"},
	{"ml","*"},
	{"dv","/"},
	{"md","%"},
	{"er","^"},
	{"ad","&"},
	{"or","|"},
	{"co","~"},
	{"nt","!"},
	{"as","="},
	{"lt","<"},
	{"gt",">"},
	{"apl","+="},
	{"ami","-="},
	{"amu","*="},
	{"adv","/="},
	{"amd","%="},
	{"aer","^="},
	{"aad","&="},
	{"aor","|="},
	{"ls","<<"},
	{"rs",">>"},
	{"ars",">>="},
	{"als","<<="},
	{"eq","=="},
	{"ne","!="},
	{"le","<="},
	{"ge",">="},
	{"aa","&&"},
	{"oo","||"},
	{"pp","++"},
	{"mm","--"},
	{"cl","()"},
	{"vc","[]"},
	{"rf","->"},
	{"cm",","},
	{"rm","->*"},
	{"op","3"}};


/*
	unmangle_str

	This function accepts a mangled C++ name and returns a simplified
	unmangled name. Returns NULL if no unmangling required.

	Note this returns a malloced buffer which should be freed when
	finished with if conversion takes place.
*/

char	*unmangle_str(char *str)
{
	char	*result=NULL;
	char	*name,*s,*s1,*cp,*op,*tp;
	unsigned len,i;

	name = malloc(strlen(str)+1);					   /*  copy for safety  */
	strcpy(name,str);
	s1 = strstr(name,"__");									  /*  get '__'  */
	if (s1==name) s1 = strstr(s1+2,"__");		 /*  ignore starting value  */
	if (s1) {									/*  found and not at start  */
		*s1 = '\0';								/*  terminate string early  */
		cp = s1+2;								 /*  search for class name  */
		len = 0;
		while ((*cp>='0')&&(*cp<='9')) {	  /*  get length of class name  */
			len *= 10;
			len += (*(cp++)-'0');
		}
		if (len) {									  /*  class name found  */
			cp[len] = '\0';								  /*  mark for use  */
		}
		else cp = NULL;								/*  else no class name  */
		op = NULL;								  /*  show no operator yet  */
		s = name;
		if (*s=='_') {						   /*  check for special names  */
			s++;
			if (s[0]=='_'&&s[1]=='_') {			  /*  need to search table  */
				for (i=0;i<(sizeof(um_table)/sizeof(struct unmangle_s));i++) {
					if (strncmp(um_table[i].name,s+2,strlen(um_table[i].name))==0) {
						op = um_table[i].oper;
						break;
					}
				}
			}
		}
		if (cp) {			  /*  if class pointer else use shortened name  */
			result = malloc(256);					/*  get largish buffer  */
			s1 = result;
			if (strcmp(name,"_base")==0) {		   /*  detect base classes  */
				s = cp;								 /*  make name = class  */
			}
			else if (name[0]=='_') *s1++ = '_';	  /*  copy '_' for sorting  */
			sprintf(s1,"%s::",cp);
			while (*s1) s1++;
			if (op) {
				switch (*op) {
				  case '1':								   /*  constructor  */
					strcpy(s1,cp);
					break;
				  case '2':									/*  destructor  */
					sprintf(s1,"~%s",cp);
					break;
				  default:									  /*  operator  */
					sprintf(s1,"operator%s",(*op=='3')?" <user>":op);
					break;
				}
			}
			else strcat(result,s);					/*  add on actual name  */
			result = realloc(result,strlen(result)+1);	 /*  adjust buffer  */
		}
		else if (op) {
			result = malloc(256);					/*  get largish buffer  */
			switch (*op) {
			  case '1':								   /*  constructor  */
			  case '2':									/*  destructor  */
				break;
			  default:									  /*  operator  */
				sprintf(result,"operator%s",(*op=='3')?" <user>":op);
				break;
			}
			result = realloc(result,strlen(result)+1);	 /*  adjust buffer  */
		}
	}
	free(name);											  /*  release copy  */
	return result;
}

/*
	unmangle_struct

	Structures need a bit of help. First unmangle all required strings and
	then pass the umangled structure member strings to this function. The
	first string is	the class name, the second string is a member name.
	The function reurns	a pointer to a buffer with the structure name in a
	more readable form if possible else NULL if the name is unchanged.

	Note this returns a malloced buffer which should be freed when
	finished with if conversion takes place.
*/

char *unmangle_struct(char *sname,char *mname)
{
	unsigned len,clen;
	char	*p=NULL,*s;

	s = malloc(strlen(sname)+3);
	sprintf(s,"%s::",sname);					   /*  get full class name  */
	len = strlen(s);
	if (strncmp(s,mname,len)==0) {			 /*  matching class in member?  */
		mname += len;									  /*  get old name  */
		p = malloc(strlen(mname)+1);						   /*  copy it  */
		strcpy(p,mname);
		free(mname);
	}
	else {										   /*  detect base classes  */
		if ((p=strstr(mname,"::"))!=NULL) {			  /*  find member name  */
			clen = p-mname;								  /*  class length  */
			if (strncmp(mname,p+2,clen)==0) {			 /*  if base class  */
				p = malloc(clen+1);							   /*  copy it  */
				strncpy(p,mname,clen);
				p[clen] = '\0';
				free(mname);
			}
		}
	}
	free(s);
	return p;
}


